from enum import Enum


class JSONAPIVersion(Enum):
    JSONAPI_RC2 = "1.0RC2"
    JSONAPI_RC3 = "1.0RC3"
    JSONAPI_1_0 = "1.0"


JSONAPI_VALID_VERSIONS = [version for version in JSONAPIVersion]
_JSONAPI_DEFAULT_VERSION = JSONAPIVersion.JSONAPI_1_0


def get_default_version():
    return _JSONAPI_DEFAULT_VERSION


def set_default_version(version):
    global _JSONAPI_DEFAULT_VERSION
    _JSONAPI_DEFAULT_VERSION = version



from cartographer.serializers import JSONAPISerializer, JSONAPINullSerializer, \
    JSONAPICollectionSerializer
from nose.tools import *


class ExampleSerializer(JSONAPISerializer):
    def __init__(self, id):
        self._id = id

    def resource_type(self):
        return "example"

    def resource_id(self):
        return str(self._id)

    def resource_url(self):
        return "http://www.example.com/examples/{}".format(
            self.resource_id(),
        )

    def attributes_dictionary(self):
        return {
            "title": "an example"
        }


class LinkingResource(JSONAPISerializer):
    def __init__(self, id, link):
        self._id = id
        self._link = link

    def resource_type(self):
        return "linking_example"

    def resource_id(self):
        return str(self._id)

    def attributes_dictionary(self):
        return {
            "title": "an example of linking"
        }

    def linked_resources(self):
        return {
            "something": self._link
        }


def test_api_resource():
    resource = ExampleSerializer(1)
    expected_json = {
        "data": {
            "type": "example",
            "id": "1",
            "attributes": {
                "title": "an example"
            }
        },
        "links": {
            "self": "http://www.example.com/examples/1",
        }
    }
    assert_equal(expected_json, resource.as_json_api_document())


def test_linked_resource():
    resource = LinkingResource(1, ExampleSerializer(2))
    expected_json = {
        "data": {
            "type": "linking_example",
            "id": "1",
            "attributes": {
                "title": "an example of linking"
            },
            "relationships": {
                "something": {
                    "data": {"type": "example", "id": "2"},
                    "links": {
                        "related": "http://www.example.com/examples/2",
                    },
                }
            }
        },
        "included": [
            {
                "type": "example",
                "id": "2",
                "attributes": {
                    "title": "an example"
                }
            }
        ]
    }
    assert_equal(expected_json, resource.as_json_api_document())


def test_linked_null():
    resource = LinkingResource(1, JSONAPINullSerializer())
    expected_json = {
        "data": {
            "type": "linking_example",
            "id": "1",
            "attributes": {
                "title": "an example of linking"
            },
            "relationships": {
                "something": {
                    "data": None
                }
            }
        }
    }
    assert_equal(expected_json, resource.as_json_api_document())


def test_linked_collection():
    resource = LinkingResource(1, JSONAPICollectionSerializer([ExampleSerializer(2), ExampleSerializer(3)]))
    expected_json = {
        "data": {
            "type": "linking_example",
            "id": "1",
            "attributes": {
                "title": "an example of linking"
            },
            "relationships": {
                "something": {
                    "data": [{"type": "example", "id": "2"}, {"type": "example", "id": "3"}]
                }
            }
        },
        "included": [
            {
                "type": "example",
                "id": "2",
                "attributes": {
                    "title": "an example"
                }
            },
            {
                "type": "example",
                "id": "3",
                "attributes": {
                    "title": "an example"
                }
            },
        ]
    }
    assert_equal(expected_json, resource.as_json_api_document())


def test_linked_heterogeneous_collection():
    resource = LinkingResource(1, JSONAPICollectionSerializer([ExampleSerializer(2), LinkingResource(2, ExampleSerializer(3))]))
    expected_json = {
        "data": {
            "type": "linking_example",
            "id": "1",
            "attributes": {
                "title": "an example of linking"
            },
            "relationships": {
                "something": {
                    "data": [{
                        "type": "example",
                        "id": "2"
                    }, {
                        "type": "linking_example",
                        "id": "2"
                    }]
                }
            }
        },
        "included": [
            {
                "type": "example",
                "id": "2",
                "attributes": {
                    "title": "an example"
                }
            },
            {
                "type": "linking_example",
                "id": "2",
                "attributes": {
                    "title": "an example of linking"
                },
                "relationships": {
                    "something": {
                        "data": {"type": "example", "id": "3"},
                        "links": {
                            "related": "http://www.example.com/examples/3",
                        },
                    },
                },
            },
            {
                "type": "example",
                "id": "3",
                "attributes": {
                    "title": "an example"
                },
            },
        ]
    }
    assert_equal.__self__.maxDiff = None
    assert_equal(expected_json, resource.as_json_api_document())


def test_toplevel_collection():
    resource = JSONAPICollectionSerializer([ExampleSerializer(2), LinkingResource(2, ExampleSerializer(3))])
    expected_json = {
        "data": [
            {
                "type": "example",
                "id": "2",
                "attributes": {
                    "title": "an example"
                }
            },
            {
                "type": "linking_example",
                "id": "2",
                "attributes": {
                    "title": "an example of linking"
                },
                "relationships": {
                    "something": {
                        "data": {"type": "example", "id": "3"},
                        "links": {
                            "related": "http://www.example.com/examples/3",
                        },
                    },
                },
            },
        ],
        "included": [
            {
                "type": "example",
                "id": "3",
                "attributes": {
                    "title": "an example"
                }
            },
        ]
    }
    assert_equal.__self__.maxDiff = None
    assert_equal(expected_json, resource.as_json_api_document())


def test_toplevel_collection_with_repeats():
    example_resource = ExampleSerializer(10)
    resource = JSONAPICollectionSerializer([
        example_resource,
        LinkingResource(2,
            JSONAPICollectionSerializer([
                example_resource,
                LinkingResource(3, example_resource)
            ])
        )
    ])
    expected_json = {
        "data": [
            {
                "type": "example",
                "id": "10",
                "attributes": {
                    "title": "an example"
                }
            },
            {
                "type": "linking_example",
                "id": "2",
                "attributes": {
                    "title": "an example of linking"
                },
                "relationships": {
                    "something": {"data": [
                        {"type": "example", "id": "10"},
                        {"type": "linking_example", "id": "3"}
                    ]}
                },
            },
        ],
        "included": [
            {
                "type": "linking_example",
                "id": "3",
                "attributes": {
                    "title": "an example of linking"
                },
                "relationships": {
                    "something": {
                        "data": {"type": "example", "id": "10"},
                        "links": {
                            "related": "http://www.example.com/examples/10",
                        },
                    },
                },
            },
        ]
    }
    assert_equal(expected_json, resource.as_json_api_document())



from cartographer.serializers import JSONAPISerializer
from cartographer.utils.version import JSONAPIVersion


class JSONAPINullSerializer(JSONAPISerializer):
    def as_linkage_json(self):
        return None

    def as_link_json(self, version):
        if version == JSONAPIVersion.JSONAPI_RC2:
            return {"id": None}
        elif version == JSONAPIVersion.JSONAPI_RC3:
            return {"linkage": self.as_linkage_json()}
        elif version == JSONAPIVersion.JSONAPI_1_0:
            return {"data": self.as_linkage_json()}
        else:
            raise ValueError("Unknown JSON API version")



from cartographer.requests.jsonapi_flask_request_mixin import JSONAPIFlaskRequestMixin
from nose.tools import *


def test_parse_parameters_to_dictionary():
    params = {
        'fields[user]': 'fields-user',
        'fields[post][one]': 'fields-post-one',
        'includes': 'users,posts'
    }
    expected_result = {
        'user': 'fields-user',
        'post': {
            'one': 'fields-post-one'
        }
    }
    assert_equals(JSONAPIFlaskRequestMixin._parse_parameters_to_dictionary(params).get('fields'),
                  expected_result)



class SchemaAttribute(object):
    """
    `SchemaAttribute` describes how to translate attributes between JSON API and our Python models.

    `SchemaAttribute` is a very simple class, with two primary methods:
    `value`, for serializing from Python into JSON,
    and `from_json` for parsing from JSON into Python.
    Subclasses of `SchemaAttribute` can override either of these methods
    to customize (de)serialization behavior, typically per JSON output type
    (string, int, bool, date, etc.)
    """

    def __init__(self, model_attribute=None, is_property=True,
                 serializer_method=None):
        """
        Note: only one of model_attribute and serializer_method should be provided.
        Note: is_property is not relevant for serializer_method

        :param model_attribute: the name of the attribute on the python model that should be (de)serialized
        :param is_property: whether or not model_attribute is a property (if False, it is presumed it is a method)
        :param serializer_method: the name of the method on the serializer object which uses this schema
        which should be called to get the serialized value.
        :return: an instance of SchemaAttribute,
        which will later be used by a Parser or Serializer to move between JSON API and Python
        """

        identifier_args = [model_attribute, serializer_method]
        provided_identifiers = [identifier
                                for identifier in identifier_args
                                if identifier]
        if len(provided_identifiers) > 1:
            raise Exception("only one of [{}] should be provided".format(identifier_args.join(", ")))

        self.model_attribute = model_attribute
        self.is_property = is_property
        self.serializer_method = serializer_method

    def to_json(self, serializer):
        """
        :param serializer: The serializer which is currently responsible for serializing this value
        :return: The value which the json library can serialize,
        which will have been formatted via format_value_for_json
        """
        value = None
        if self.model_attribute:
            value = getattr(serializer.model, self.model_attribute)
            if not self.is_property:
                value = value()
        elif self.serializer_method:
            value = getattr(serializer, self.serializer_method)()

        if value is not None:
            value = self.format_value_for_json(value)

        return value

    @classmethod
    def format_value_for_json(cls, value):
        """
        :param value: The raw value, almost always the output of self.value
        :return: The formatted value, which can be dropped into a Python dictionary as the output of this Serializer,
        typically which will have json.dumps called on it to return to a client
        """
        return value

    @classmethod
    def from_json(cls, serialized_value):
        """
        :param serialized_value: The value to be parsed, as pulled directly from RequestInterface.get_json(force=True)
        :return: The parsed value, which can be dropped into a Python dictionary as the output of this Parser,
        typically which will be inserted into a SQLAlchemy table.
        """
        return serialized_value



from cartographer.field_types import SchemaAttribute


class BoolAttribute(SchemaAttribute):
    @classmethod
    def format_value_for_json(cls, value):
        return bool(value)

    @classmethod
    def from_json(cls, serialized_value):
        return bool(str(serialized_value).lower() not in ['false', '0'])



from cartographer.field_types import SchemaAttribute


class JSONAttribute(SchemaAttribute):
    @classmethod
    def format_value(cls, value):
        # the final serializer will appropriately dump to json, we should *not* do that prematurely here
        return value



from abc import ABCMeta, abstractmethod


class JSONAPIRequestInterface(object):
    _metaclass__ = ABCMeta

    @abstractmethod
    def get_json_api_version(self):
        return

    @abstractmethod
    def get_json(self, force=False):
        return

    @abstractmethod
    def get_includes(self):
        return

    @abstractmethod
    def get_requested_fields(self):
        return



from cartographer.field_types import SchemaRelationship
from cartographer.resources.resource_registry import ResourceRegistryKeys


class ArrayRelationship(SchemaRelationship):
    def __init__(self, model_type, model_property=None, model_method=None,
                 serializer_method=None, includes=None):
        super().__init__(model_type=model_type, model_property=model_property, model_method=model_method,
                         serializer_method=serializer_method, includes=includes)

    def related_serializer(self, parent_serializer, relationship_key):
        from cartographer.serializers import JSONAPICollectionSerializer

        if self.serializer_method is not None:
            return getattr(parent_serializer, self.serializer_method)()

        models = []
        if self.id_attribute is not None:
            raise Exception("id_attribute provided for an ArrayRelationship.")
        elif self.model_property is not None:
            models = getattr(parent_serializer.model, self.model_property)
        elif self.model_method is not None:
            models = getattr(parent_serializer.model, self.model_method)()

        serializer_class = self.resource_registry_entry().get(ResourceRegistryKeys.SERIALIZER)
        # TODO: custom collection class, custom arguments to serializer_class
        return JSONAPICollectionSerializer([
            serializer_class(
                model,
                parent_serializer=parent_serializer,
                relationship_name=relationship_key
            )
            for model in models
        ])



from datetime import timezone as timezone_


def as_utc(dt):
    if hasattr(dt, 'tzinfo'):
        if dt.tzinfo:
            return dt.astimezone(timezone_.utc)
        else:
            return dt.replace(tzinfo=timezone_.utc)
    return dt


def make_naive(dt):
    return as_utc(dt).replace(tzinfo=None)



from cartographer.parsers.jsonapi_parser import PostedDocument
from cartographer.field_types import ArrayRelationship
from cartographer.utils.version import get_default_version


class SchemaParser(PostedDocument):
    """
    SchemaParser turns JSON API inputs into python dictionaries
    suitable to pass as arguments to SQLAlchemy `.insert()` and `.update()` calls
    (hereafter referred to as "table data").
    It does so by subclassing `PostedDocument` and using those convenience methods
    to provide validation and deserialization.
    """

    def __init__(self, json_data=None,
                 inbound_request=None, inbound_session=None,
                 parent_parser=None, relationship_name=None,
                 current_user_id=None,
                 version=None):
        """
        Route files should pass in `inbound_request` and `inbound_session`,
        Resource files should pass in `parent_parser` and `relationship_name`

        :param json_data: The json document which will be parsed into table data
        :param inbound_request: The current `JSONAPIRequest` which is prompting the parsing of this resource
        :param inbound_session: The `JSONAPISession` which is active during the parsing of this resource
        :param parent_parser: The `PostedDocument` which is creating this instance for assistance in parsing one of its related resources
        :param relationship_name: The name by which the parent_parser refers to this instance
        :param current_user_id: The ID of the user on behalf of whom the resource is being parsed
        :param version: The version of the JSON API spec that the posted document is using

        :return: A instance of SchemaParser (or more typically a subclass),
        which will typically have `validated_table_data` called on it in a route
        to extract usable data for our python models.
        """

        if version is None:
            if inbound_request:
                version = inbound_request.get_json_api_version()
            else:
                version = get_default_version()
        if json_data is None and inbound_request:
            json_data = inbound_request.get_json(force=True)
        super().__init__(json_data=json_data, version=version)

        if parent_parser and not current_user_id:
            current_user_id = parent_parser.current_user_id

        if current_user_id is None and inbound_session:
            current_user_id = inbound_session.user_id
        self.current_user_id = str(current_user_id)

        self._table_data = None

    @classmethod
    def schema(cls):
        """Override this in a subclass to define model <=> API mappings"""
        raise NotImplementedError()

    def validated_table_data(self):
        """The primary use case for this class. Simply calls validate, then returns the table data"""
        self.validate(self.data())
        return self.table_data()

    # # Parsing

    def table_data(self):
        """
        The primary entry point for parsing.

        :return: A map from column names to their values
        """
        if self._table_data is None:
            result = {}

            data = self.data()

            json_attributes = data.json_data.get('attributes', {})
            if json_attributes:
                for key in json_attributes:
                    if self.should_parse_attribute(key):
                        parsed_attribute = self.parse_schema_attribute(key)
                        if parsed_attribute and len(parsed_attribute) == 2:
                            result_key, parsed_value = parsed_attribute
                            result[result_key] = parsed_value

            json_relationships = data.json_data.get('relationships', {})
            if json_relationships:
                for key in json_relationships:
                    if self.should_parse_relationship(key):
                        parsed_relationship = self.parse_schema_relationship(key)
                        if parsed_relationship and len(parsed_relationship) == 2:
                            result_key, parsed_value = parsed_relationship
                            result[result_key] = parsed_value

            self._table_data = result
        return self._table_data

    # Attributes

    def parse_schema_attribute(self, key):
        """
        Note: currently only handles SchemaAttributes which are model properties.
        The assumption here is that computed values are read-only.

        :param key: The name of the attribute to be parsed
        :return: A tuple of the model property name and its corresponding parsed value
        """
        schema_attribute = self.schema().attribute(key)
        if schema_attribute \
                and schema_attribute.model_attribute \
                and schema_attribute.is_property:
            serialized_value = self.data().attribute(key)
            return schema_attribute.model_attribute, schema_attribute.from_json(serialized_value)
        # TODO: let SchemaAttribute declare parser_method

    def should_parse_attribute(self, key):
        return key in self.schema().attributes()

    # Relationships

    def parse_schema_relationship(self, key):
        """
        Note: currently only handles SchemaRelationships which are map to foreign keys.
        This is because rich objects are not `insert`-able or `update`-able,
        and the return format is column values.

        :param key: The name of the relationship to be parsed
        :return: A tuple of the model property name and its corresponding parsed value.
        Note: for to-many relationships, the return value is formatted as a tuple with
        * the relationship key
        * an array of maps from 'id' *for the other table* to corresponding parsed values.
        """
        schema_relationship = self.schema().relationship(key)
        if schema_relationship:
            # TODO: redo array relationship parsing.
            # there was a thing where everything was nested under keys. probably that.
            if isinstance(schema_relationship, ArrayRelationship):
                return key, [
                    {'id': x.resource_id()}
                    for x in self.data().relationship_id(key)
                ]
            elif schema_relationship.id_attribute:
                return schema_relationship.id_attribute, self.data().relationship_id(key).resource_id()
        return None
        # TODO: let SchemaRelationship declare parser_method

    def should_parse_relationship(self, key):
        return key in self.schema().relationships()

    # # Validation

    def validate(self, inbound_data):
        """Override this in a subclass to enforce format and content rules for the inbound JSON API document"""
        pass



from cartographer.resources import get_resource_registry
from cartographer.resources.resource_registry import ResourceRegistryKeys


class SchemaRelationship(object):
    """
    `SchemaRelationship` describes how to translate related resources to and from JSON API and our Python models.

    `SchemaRelationship` is has one primary method,
    `related_serializer`, for creating a `JSONAPISerializer` instance based on its input arguments.
    Subclasses of `SchemaSerializer` can override this method
    to customize serialization behavior.
    Parsing of related resources is not currently handled by this class,
    and instead is handled by the `PostedDocument` class (or, more typically, its subclass `SchemaParser`.
    """

    def __init__(self, model_type, id_attribute=None, model_property=None,
                 model_method=None, serializer_method=None, includes=None):
        """
        NOTE: only one of id_attribute, model_property, model_method, or serializer_method should be provided

        :param model_type: the JSON API `type` string for the related model
        :param id_attribute: the foreign key column on the parent serializer model which identifies the related serializer
        :param model_property: the property on the parent serializer model which returns the related serializer
        :param model_method: the property on the parent serializer model which returns the related serializer
        :param serializer_method: the name of the method on the parent serializer object which uses this schema
        which should be called to get the child serializer.
        :return: an instance of SchemaRelationship,
        which will later be used to serialize Python into JSON API.
        """

        identifier_args = [id_attribute, model_property, model_method, serializer_method]
        provided_identifiers = [identifier
                                for identifier in identifier_args
                                if identifier]
        if len(provided_identifiers) > 1:
            raise Exception("only one of [{}] should be provided".format(identifier_args.join(", ")))

        self.model_type = model_type
        self.id_attribute = id_attribute
        self.model_property = model_property
        self.model_method = model_method
        self.serializer_method = serializer_method
        self.includes = includes

    def related_serializer(self, parent_serializer, relationship_key):
        """
        :param parent_serializer: The serializer which has our return value as a related resource
        :param relationship_key: The name by which the parent serializer knows this child
        :return: The child serializer which will later be used to serialize a related resource
        """
        if self.serializer_method is not None:
            return getattr(parent_serializer, self.serializer_method)()

        model = None
        if self.id_attribute is not None:
            related_model_getter = self.resource_registry_entry().get(ResourceRegistryKeys.MODEL_GET)
            model_id = getattr(parent_serializer.model, self.id_attribute)
            if model_id is not None and related_model_getter is not None:
                model = related_model_getter(model_id)
        elif self.model_property is not None:
            model = getattr(parent_serializer.model, self.model_property)
        elif self.model_method is not None:
            model = getattr(parent_serializer.model, self.model_method)()

        if model:
            serializer_class = self.resource_registry_entry().get(ResourceRegistryKeys.SERIALIZER)
            return serializer_class(
                model,
                parent_serializer=parent_serializer,
                relationship_name=relationship_key,
                includes=self.includes
            )
        else:
            from cartographer.serializers import JSONAPINullSerializer
            return JSONAPINullSerializer()

    def resource_registry_entry(self):
        return get_resource_registry().get(self.model_type, {})



from .resource_registry import ResourceRegistry

_resource_registry = ResourceRegistry()


def get_resource_registry_container():
    global _resource_registry
    _resource_registry.call_initialization_hook()
    return _resource_registry


def get_resource_registry():
    return get_resource_registry_container().registry



from cartographer.field_types import ArrayRelationship, SchemaRelationship
from cartographer.permissions.base_mask import BaseMask
from cartographer.resources import get_resource_registry
from cartographer.resources.resource_registry import ResourceRegistryKeys
from cartographer.serializers import JSONAPISerializer
from cartographer.utils import config


class SchemaSerializer(JSONAPISerializer):
    """
    SchemaSerializer turns python models into JSON API responses.
    It does so by subclassing `JSONAPISerializer`, and implementing its main four methods --
    `resource_id`, `resource_type`, `attributes_dictionary`, and `linked_resources` --
    using the passed-in model and the associated `schema`.
    """

    def __init__(self, model,
                 inbound_request=None, inbound_session=None,
                 parent_serializer=None, relationship_name=None,
                 includes=None, requested_fields=None,
                 current_user_id=None):
        """
        Route files should pass in `inbound_request` and `inbound_session`,
        Resource files should pass in `parent_serializer` and `relationship_name`

        :param model: The model which will be serialized into JSON
        :param inbound_request: The current `JSONAPIRequest` which is prompting the creation of this resource
        :param inbound_session: The `JSONAPISession` which is active during the creation of this resource
        :param parent_serializer: The `JSONAPISerializer` which is creating this instance as one of its `linked_resources`
        :param relationship_name: The name by which the parent_serializer refers to this instance
        :param includes: An array of strings, representing the relationships which should be serialized
        :param requested_fields: A map from strings to arrays of strings,
            representing the resource types and their associated attributes which should be serialized
        :param current_user_id: The ID of the user on behalf of whom the resource is being created

        :return: A instance of SchemaSerializer (or more typically a subclass),
        which will typically have `as_json_api_document` called on it in a route to transform it into a JSON response.
        """

        self.model = model

        if parent_serializer:
            if not requested_fields:
                requested_fields = parent_serializer.requested_fields
            if not current_user_id:
                current_user_id = parent_serializer.current_user_id

        if includes is None:
            if parent_serializer is not None and relationship_name is not None:
                prefix = relationship_name + '.'
                includes = [include_path[len(prefix):]
                            for include_path in parent_serializer.includes
                            if include_path.startswith(prefix)]
                if len(includes) == 0:
                    # TODO: kill all uses of request.args.get('use-defaults-for-included-resources') in clients
                    includes = type(self).default_includes()
            else:
                if inbound_request:
                    includes = inbound_request.get_includes()
                if includes is None:
                    includes = type(self).default_includes()
        self.includes = includes

        if requested_fields is None and inbound_request:
            requested_fields = inbound_request.get_requested_fields()
        self.requested_fields = requested_fields

        if current_user_id is None and inbound_session:
            current_user_id = inbound_session.user_id
        self.current_user_id = current_user_id

        self._linked_resources = None
        self._masked_fields = None
        self._masked_includes = None

        self.prime_for_includes()

    @classmethod
    def schema(cls):
        """Override this in a subclass to define model <=> API mappings"""
        raise NotImplementedError()

    @classmethod
    def default_fields(cls):
        """Override this in a subclass to define which attributes should be included by default"""
        return cls.schema().attributes()

    @classmethod
    def default_includes(cls):
        """Override this in a subclass to define which relationships should be included by default"""
        return cls.schema().relationships()

    @classmethod
    def model_class(cls):
        return get_resource_registry().get(cls.resource_type(), {}).get(ResourceRegistryKeys.MODEL)

    @classmethod
    def serializer_class(cls):
        return get_resource_registry().get(cls.resource_type(), {}).get(ResourceRegistryKeys.SERIALIZER)

    @classmethod
    def mask_class(cls):
        return get_resource_registry().get(cls.resource_type(), {}).get(ResourceRegistryKeys.MASK, BaseMask)

    @classmethod
    def resource_type(cls):
        return cls.schema().resource_type()

    def resource_id(self):
        return self.schema().resource_id().to_json(self)

    @classmethod
    def route_prefix(cls):
        return cls.schema().route_prefix()

    def resource_url(self):
        if config.api_server and self.route_prefix():
            return "https://{server}/{route}/{id}".format(
                server=config.api_server,
                route=self.route_prefix(),
                id=self.resource_id_str()
            )
        return super().resource_url()

    # # Serialization

    # Included resources

    def linked_resources(self):
        """
        The primary entry point for serialization of relationships

        :return: A map from relationship names to the related resource
        """
        if self._linked_resources is None:
            links = {}
            relationship_keys = self.schema().relationships()
            if relationship_keys:
                for key in relationship_keys:
                    if self.should_include_relationship(key):
                        relationship = self.schema().relationship(key)
                        links[key] = relationship.related_serializer(self, key)
            self._linked_resources = links
        return self._linked_resources

    def normalized_includes(self):
        # If includes passed in are e.g ['pledges.campaign.goals', 'campaign'] for a UserResource,
        # we want to include 'pledges' and 'campaign' here,
        # even tho the literal string 'pledges' is not in the self.includes array
        return [include.split('.')[0]
                for include in self.includes]

    def should_include_relationship(self, key):
        """
        Checks if the user requested the related resource and our Masks allow it.

        :param key: The name by which the parent resource refers to the child resource
        :return: A boolean indicating whether or not the relationship matching the given key should be serialized
        """
        if key not in self.normalized_includes():
            return False
        if self._masked_includes is None:
            self._masked_includes = self.mask_class().includes_cant_view(self.model, self.current_user_id)
        return key not in self._masked_includes

    # SQL Priming for Serialization

    def prime_for_includes(self):
        """
        While serializing resources, a tree structure linking parent resources to child relationships is formed.
        Sibling nodes in this tree do not know about each other,
        yet often make nearly identical SQL requests to hydrate their attributes and relationships.
        Priming is how we make this more performant --
        at initialization time, each sibling node primes the queries it will perform,
        so that when a node eventually performs the query, it can perform all the queries together
        (via our `https://github.com/Patreon/flask-caching-services` `MultigetCache` wrapper)
        """
        relationship_keys = self.schema().relationships()
        if relationship_keys:
            for key in relationship_keys:
                # Often times checking should_include_relationship issues a query,
                # but priming is cheap, so we over-prime rather than check self.should_include_relationship(key)
                self.prime_schema_relationship(key)

    def prime_schema_relationship(self, key):
        relationship_data_schema = self.schema().relationship(key)
        relationship_type = None
        id_attribute = None
        if isinstance(relationship_data_schema, SchemaRelationship):
            if not isinstance(relationship_data_schema, ArrayRelationship):
                relationship_type = relationship_data_schema.model_type
                id_attribute = relationship_data_schema.id_attribute

        if relationship_type is not None and id_attribute is not None and hasattr(self.model, id_attribute):
            relationship_model_get_primer = get_resource_registry().get(relationship_type) \
                .get(ResourceRegistryKeys.MODEL_PRIME)
            if relationship_model_get_primer is not None:
                relationship_model_get_primer(getattr(self.model, id_attribute))

    # Attributes

    def attributes_dictionary(self):
        """
        The primary entry point for serialization of attributes

        :return: A map from attribute names to their values
        """
        result = {}
        attribute_keys = self.schema().attributes()
        if attribute_keys:
            for key in attribute_keys:
                if self.should_include_attribute(key):
                    attribute = self.schema().attribute(key)
                    result[key] = attribute.to_json(self)
        return result

    def should_include_attribute(self, key):
        """
        Checks if the user requested the attribute and our Masks allow it.

        :param key: The name of the attribute of the resource
        :return: A boolean indicating whether or not the attribute matching the given key should be serialized
        """
        if self.requested_fields is not None and self.resource_type() in self.requested_fields:
            if key not in self.requested_fields[self.resource_type()]:
                return False
        elif key not in self.default_fields():
            return False
        if self._masked_fields is None:
            self._masked_fields = self.mask_class().fields_cant_view(self.model, self.current_user_id)
        return key not in self._masked_fields



from cartographer.parsers.schema_parser import SchemaParser
from cartographer.permissions.base_mask import BaseMask
from cartographer.resources import get_resource_registry_container
from cartographer.resources.resource_registry import ResourceRegistryKeys
from cartographer.schemas.schema import Schema
from cartographer.serializers import SchemaSerializer


class APIResource(object):
    SCHEMA = Schema
    SERIALIZER = SchemaSerializer
    PARSER = SchemaParser
    MASK = BaseMask
    MODEL = None
    MODEL_GET = None
    MODEL_PRIME = None

    @classmethod
    def type_string(cls):
        return cls.SCHEMA.schema()['type']

    @classmethod
    def register(cls, registry_key):
        def wrapping_function(wrapped_class):
            if registry_key == ResourceRegistryKeys.SCHEMA:
                cls.SCHEMA = wrapped_class
            if registry_key == ResourceRegistryKeys.SERIALIZER:
                cls.SERIALIZER = wrapped_class
            if registry_key == ResourceRegistryKeys.PARSER:
                cls.PARSER = wrapped_class
            if registry_key == ResourceRegistryKeys.MASK:
                cls.MASK = wrapped_class
            if registry_key == ResourceRegistryKeys.MODEL:
                cls.MODEL = wrapped_class
            if registry_key == ResourceRegistryKeys.MODEL_GET:
                cls.MODEL_GET = wrapped_class
            if registry_key == ResourceRegistryKeys.MODEL_PRIME:
                cls.MODEL_PRIME = wrapped_class

            cls.register_class()

            return wrapped_class

        return wrapping_function

    @classmethod
    def register_class(cls):
        registry = get_resource_registry_container()
        registry.register_resource(
            type_string=cls.type_string(),
            schema=cls.SCHEMA,
            serializer=cls.SERIALIZER,
            parser=cls.PARSER,
            mask=cls.MASK,
            model=cls.MODEL,
            model_get=cls.MODEL_GET,
            model_prime=cls.MODEL_PRIME
        )



from setuptools import setup, find_packages

setup(name='cartographer',
      version='0.1.1',
      description='Python library for using JSON API, especially with Flask.',
      url='http://github.com/Patreon/cartographer',
      author='Patreon',
      author_email='david@patreon.com',
      license='Apache 2.0',
      packages=find_packages(exclude=['example', 'example.*', 'test', 'test.*']),
      install_requires=[
          'python-dateutil==2.4.2',
          'ciso8601==1.0.1'
      ],
      zip_safe=True,
      classifiers=[
          'Development Status :: 2 - Pre-Alpha',
          'Environment :: Web Environment',
          'Framework :: Flask',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: Apache Software License',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Topic :: Internet :: WWW/HTTP',
          'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
          'Topic :: Software Development :: Libraries :: Python Modules',
      ]
)



import json
import re

from cartographer.exceptions.request_exceptions import DataMissing, BadPageCountParameter, BadPageCursorParameter, \
    BadPageOffsetParameter
from cartographer.requests.jsonapi_request_interface import JSONAPIRequestInterface
from cartographer.utils.version import JSONAPIVersion, get_default_version


class JSONAPIFlaskRequestMixin(JSONAPIRequestInterface):
    def get_jsonapi_json(self):
        force_if_correct_mime = (self.mimetype == 'application/vnd.api+json')
        return self.get_json(force=force_if_correct_mime)

    def get_jsonapi_data(self):
        sent_json = self.get_jsonapi_json()
        if sent_json and 'data' in sent_json:
            data = sent_json.get('data')
        elif self.form and self.form.get('data'):
            data_str = self.form.get('data')
            data = json.loads(data_str)
        else:
            raise DataMissing()
        # TODO: have get_data clients expect attributes to be nested
        if self.get_json_api_version() == JSONAPIVersion.JSONAPI_1_0:
            data.update(data.get('attributes', {}))
        return data

    def get_sort(self):
        return self.args.get('sort')

    def get_pagination(self, page_count_default=10, cursor_formatter=None):
        per_page = self.args.get('page[count]')
        if per_page is not None:
            if per_page != 'infinity':
                try:
                    per_page = int(per_page)
                except TypeError:
                    raise BadPageCountParameter(per_page)
        else:
            per_page = page_count_default

        page_cursor = self.args.get('page[cursor]')
        if page_cursor is not None and cursor_formatter is not None:
            try:
                page_cursor = cursor_formatter(page_cursor)
            except:
                raise BadPageCursorParameter(page_cursor)

        page_offset = self.args.get('page[offset]')
        if page_offset is not None:
            try:
                page_offset = int(page_offset)
            except TypeError:
                raise BadPageOffsetParameter(page_offset)

        return page_cursor, page_offset, per_page

    def get_includes(self):
        """Returns a list of the requested resources to include in the response for this request."""
        includes_string = self.args.get('include')
        if includes_string:
            if includes_string in ['null', 'none', '[]']:
                return []
            else:
                return includes_string.split(',')
        else:
            return None

    def get_requested_fields(self):
        """Returns a list of the requested attributes to include in the response for this request."""
        type_to_list = self.dictionary_from_get('fields')
        return {key: val.split(',') for key, val in type_to_list.items()}

    def get_included(self):
        """Returns a list of the included resources sent from the client"""
        sent_json = self.get_jsonapi_json()
        if not sent_json:
            raise DataMissing()

        return sent_json.get('included', [])

    def get_included_with_filter(self, type_, id_=None):
        included = self.get_included()
        return [
            resource
            for resource in included
            if resource.get('type') == type_ and (id_ is None or id_ == resource.get('id'))
        ]

    def get_filters(self):
        return self.dictionary_from_get('filter')

    def get_bool(self, key):
        val = self.args.get(key)
        if val is None:
            return None
        return bool(str(val).lower() not in ['false', '0'])

    def get_json_api_version(self, default_version=None):
        if default_version is None:
            default_version = get_default_version()
        return JSONAPIVersion(self.args.get('json-api-version', default_version))

    def dictionary_from_get(self, outer_key):
        return self._parse_parameters_to_dictionary(self.args).get(outer_key, {})

    @staticmethod
    def _parse_parameters_to_dictionary(params):
        pattern = re.compile('([^\[\]]+)\[(.*)\]')
        return_dict = {}

        for key in params.keys():
            match_data = pattern.fullmatch(key)
            if not match_data:
                continue

            match_groups = match_data.groups()
            if len(match_groups) != 2:
                continue

            prefix = match_groups[0]
            path = match_groups[1].split('][')
            result_for_prefix = return_dict.get(prefix, {})
            current_result = result_for_prefix
            for index, step in enumerate(path):
                if index == len(path) - 1:
                    current_result[step] = params[key]
                else:
                    if step not in current_result:
                        current_result[step] = {}
                    current_result = current_result[step]
            return_dict[prefix] = result_for_prefix

        return return_dict
